function demo_iterativeQMIS
% This program implements 'iterative QMI-S' (a domain adaptation algorithm for object class recognition problem).

% ref: DOMAIN ADAPTATION BY ITERATIVE IMPROVEMENT OF SOFT-LABELING AND MAXIMIZATION OF NON-PARAMETRIC MUTUAL INFORMATION.  
% M.N.A. Khan, Douglas R. Heisterkamp.  
% Proceedings of the IEEE International Conference on Image Processing (ICIP), Phoenix, Arizona, Sept 2016.

%  
% version 2.0 --Jul/2016
% version 1.0 --Jan/2016
% Written by: M. N. A. Khan (mohk@cs.okstate.edu)

k = 9;             % dimension size of the learned subspace     
loop = 200;        % iteration count
subprob = 1;       % subproblem number
thresh = 95;       % threshold for PCA whitening

source = {'Caltech10','Caltech10','Caltech10','amazon','amazon','amazon','webcam','webcam','webcam','dslr','dslr','dslr'}; 
target = {'amazon','webcam','dslr','Caltech10','webcam','dslr','Caltech10','amazon','dslr','Caltech10','amazon','webcam'};
src = char(source{subprob});
tgt = char(target{subprob});
    
%% Preprocess data using z-scoring. Data is stored in row-major ordering i.e. each row represents one feature representation.
load(['./data/' src '_SURF_L10.mat']);
fts = fts ./ repmat(sum(fts,2),1,size(fts,2));
Xs = zscore(fts);
src_count = size(Xs,1);
Ys = labels;

load(['./data/' tgt '_SURF_L10.mat']);
fts = fts ./ repmat(sum(fts,2),1,size(fts,2)); 
Xt = zscore(fts);
trg_count = size(Xt,1);
Yt = labels;

total_count = src_count+trg_count;
nClasses = numel(unique(Ys));
    
% Data whitening by PCA (preserving 95% variance)
X = [Xs;Xt];
[U,~,dd] = princomp(X);  
cc = cumsum(dd)/sum(dd); 
idx = find(cc>=thresh/100); 
dim=idx(1);
X = X*U(:,1:dim);


%% Computing sigma and centralized Kernel matrix
dis_array=zeros(1,total_count*(total_count-1)/2);
Dis = EuDist2(X,[],0);
fsig=0;ct = 0;
for i1=1:size(Dis,1)
   for i2=1:size(Dis,2)
     if i1>i2
        fsig = fsig + Dis(i1,i2);
        ct = ct + 1;
        dis_array(ct) = Dis(i1,i2);
     end
   end
end
sigma_k = median(dis_array);

En = (1/total_count)*ones(total_count);
D = EuDist2(X,[],0);
K_t = exp(-D/(2*sigma_k^2));
K_t = max(K_t,K_t');
K = K_t - En*K_t - K_t*En + En*K_t*En;  
[Zk,Dk] = svd(K);
KHalf = Zk*(Dk.^.5)*Zk';

knn = round(log(src_count)) + 1;

% Computing classification performance in original feature space wothout
% adaptation
data = strcat(src,'(',num2str(src_count),')',' ->  ',tgt,'(',num2str(trg_count),')');
mdl = fitcknn(Xs,Ys,'NumNeighbors',1);
[l1,~,~] = predict(mdl,Xt);
accu = length(find(l1==Yt))/length(Yt);

fprintf(['Subproblem %d: ' data  '   knn=%d   sigma=%.2f   \n'],subprob,knn,sigma_k);
fprintf('Accuracy in orginal feature space without adaptation = %f%% \n',accu*100);


%% Initialize soft-labeling for data; labels are known for source data, whereas target data
%labels are unknown and hence initialized with uniform class distribution

score_all = zeros(total_count,numel(unique(Ys)));
for cc=1:src_count
    score_all(cc,Ys(cc)) = 1;     
end
for cc=1+src_count:total_count
    score_all(cc,:) = ones(1,nClasses)/nClasses;     
end
    
for t=1:loop
    [base2] = KQMI(nClasses,score_all,k,total_count);
    Z = KHalf * base2;        % projected data in learned subspace
   
    if t>1
       prev_theta = cur_theta;
       cur_theta = subspace(base1,base2);
    else
       cur_theta = 0;
    end
    base1 = base2;
    
    % updating label prediction of target data
    Zs = Z(1:src_count,:);  
    Zt = Z(1+src_count:total_count,:);  
    mdl = fitcknn(Zs,Ys,'NumNeighbors',knn);
    [lab,sc,~] = predict(mdl,Zt);
    accu = length(find(lab==Yt))/length(Yt);  
    score_all(1+src_count:total_count,:) = sc;
    
    fprintf('%d: acc=%.2f%%  sub=%.4f \n'   ,t,accu*100, cur_theta);
    
    if t>2 && abs(cur_theta-prev_theta)<=1e-4
        break;
    end
end 

%% Final classification by 1-NN classifier
mdl = fitcknn(Zs,Ys,'NumNeighbors',1);
[lab,~,~] = predict(mdl,Zt);
accuf = length(find(lab==Yt))/length(Yt);
fprintf('\n\n classification performance in percent: %.2f \n',accuf*100);
     
end

%% This function computes the basis of learned subspace
function [base] = KQMI(n_classes,score,k,n)
    
    Sc=zeros(n_classes,1);
    for c=1:n_classes
        Sc(c) = sum(score(:,c))/n;
    end
    
    % computing V_in
    V_in = zeros(n);
    for c=1:10
        V_in = V_in + score(:,c)*score(:,c)';
    end
    V_in = V_in/(n^2);
    
    % computing M
    on = ones(n,1);
    temp = sum(Sc.^2);
    V_all = (on*on')*(temp/(n^2));
        
    % computing V_BTW
    q = zeros(1,n);
    for c=1:10
        one_c = score(:,c);
        q = q + (Sc(c)/(n^2)) * one_c';  
    end
    V_btw = -2*on*q;
 
    M = V_in+V_all+V_btw;
   
    M = (M + M')/2; 
    [base,~] = eigs(M,k,'LM');
end   

